@@ -8,14 +8,18 @@
     html_root_url = "https://doc.rust-lang.org/nightly/nightly-rustc/",
     test(attr(deny(warnings)))
 )]
+#![feature(associated_type_bounds)]
 #![feature(box_patterns)]
-#![feature(crate_visibility_modifier)]
+#![feature(const_default_impls)]
+#![feature(const_trait_impl)]
 #![feature(if_let_guard)]
 #![feature(label_break_value)]
-#![feature(nll)]
+#![feature(let_chains)]
 #![feature(min_specialization)]
-#![recursion_limit = "256"]
+#![feature(negative_impls)]
 #![feature(slice_internals)]
+#![feature(stmt_expr_attributes)]
+#![recursion_limit = "256"]
 
 #[macro_use]
 extern crate rustc_macros;
@@ -29,7 +33,7 @@ pub mod util {
 }
 
 pub mod ast;
-pub mod ast_like;
+pub mod ast_traits;
 pub mod attr;
 pub mod entry;
 pub mod expand;
@@ -41,7 +45,7 @@ pub mod tokenstream;
 pub mod visit;
 
 pub use self::ast::*;
-pub use self::ast_like::AstLike;
+pub use self::ast_traits::{AstDeref, AstNodeWrapper, HasAttrs, HasNodeId, HasSpan, HasTokens};
 
 use rustc_data_structures::stable_hasher::{HashStable, StableHasher};
 
