@@ -1,7 +1,6 @@
-#![feature(bool_to_option)]
-#![feature(crate_visibility_modifier)]
-#![feature(or_patterns)]
+#![feature(associated_type_bounds)]
 #![feature(box_patterns)]
+#![feature(with_negative_coherence)]
 #![recursion_limit = "256"]
 
 mod helpers;
