@@ -1,13 +1,13 @@
 //! Constants for the 32-bit signed integer type.
 //!
-//! *[See also the `i32` primitive type](../../std/primitive.i32.html).*
+//! *[See also the `i32` primitive type][i32].*
 //!
 //! New code should use the associated constants directly on the primitive type.
 
 #![stable(feature = "rust1", since = "1.0.0")]
-#![rustc_deprecated(
+#![deprecated(
     since = "TBD",
-    reason = "all constants in this module replaced by associated constants on `i32`"
+    note = "all constants in this module replaced by associated constants on `i32`"
 )]
 
 int_module! { i32 }
