@@ -5,9 +5,9 @@
 //! New code should use the associated constants directly on the primitive type.
 
 #![stable(feature = "rust1", since = "1.0.0")]
-#![rustc_deprecated(
+#![deprecated(
     since = "TBD",
-    reason = "all constants in this module replaced by associated constants on `i64`"
+    note = "all constants in this module replaced by associated constants on `i64`"
 )]
 
 int_module! { i64 }
