@@ -6,9 +6,9 @@ pub fn get_concurrency() -> usize {
     if let Ok(value) = env::var("RUST_TEST_THREADS") {
         match value.parse::<NonZeroUsize>().ok() {
             Some(n) => n.get(),
-            _ => panic!("RUST_TEST_THREADS is `{}`, should be a positive integer.", value),
+            _ => panic!("RUST_TEST_THREADS is `{value}`, should be a positive integer."),
         }
     } else {
-        thread::available_concurrency().map(|n| n.get()).unwrap_or(1)
+        thread::available_parallelism().map(|n| n.get()).unwrap_or(1)
     }
 }
