@@ -1,9 +1,9 @@
 #![feature(rustc_private)]
 
 // NOTE: For the example to compile, you will need to first run the following:
-//   rustup component add rustc-dev
+//   rustup component add rustc-dev llvm-tools-preview
 
-// version: 1.53.0-nightly (9b0edb7fd 2021-03-27)
+// version: 1.62.0-nightly (7c4b47696 2022-04-30)
 
 extern crate rustc_error_codes;
 extern crate rustc_errors;
@@ -13,13 +13,12 @@ extern crate rustc_interface;
 extern crate rustc_session;
 extern crate rustc_span;
 
+use std::{path, process, str};
+
 use rustc_errors::registry;
 use rustc_hash::{FxHashMap, FxHashSet};
-use rustc_session::config;
+use rustc_session::config::{self, CheckCfg};
 use rustc_span::source_map;
-use std::path;
-use std::process;
-use std::str;
 
 fn main() {
     let out = process::Command::new("rustc")
@@ -36,18 +35,22 @@ fn main() {
         },
         // cfg! configuration in addition to the default ones
         crate_cfg: FxHashSet::default(), // FxHashSet<(String, Option<String>)>
+        crate_check_cfg: CheckCfg::default(), // CheckCfg
         input: config::Input::Str {
-            name: source_map::FileName::Custom("main.rs".to_string()),
-            input: "static HELLO: &str = \"Hello, world!\"; fn main() { println!(\"{}\", HELLO); }"
-                .to_string(),
+            name: source_map::FileName::Custom("main.rs".into()),
+            input: r#"
+static HELLO: &str = "Hello, world!";
+fn main() {
+    println!("{HELLO}");
+}
+"#
+            .into(),
         },
         input_path: None,  // Option<PathBuf>
         output_dir: None,  // Option<PathBuf>
         output_file: None, // Option<PathBuf>
         file_loader: None, // Option<Box<dyn FileLoader + Send + Sync>>
         diagnostic_output: rustc_session::DiagnosticOutput::Default,
-        // Set to capture stderr output during compiler execution
-        stderr: None,                    // Option<Arc<Mutex<Vec<u8>>>>
         lint_caps: FxHashMap::default(), // FxHashMap<lint::LintId, lint::Level>
         // This is a callback from the driver that is called when [`ParseSess`] is created.
         parse_sess_created: None, //Option<Box<dyn FnOnce(&mut ParseSess) + Send>>
@@ -70,15 +73,17 @@ fn main() {
         compiler.enter(|queries| {
             // Parse the program and print the syntax tree.
             let parse = queries.parse().unwrap().take();
-            println!("{:#?}", parse);
+            println!("{parse:?}");
             // Analyze the program and inspect the types of definitions.
             queries.global_ctxt().unwrap().take().enter(|tcx| {
-                for item in tcx.hir().items() {
+                for id in tcx.hir().items() {
+                    let hir = tcx.hir();
+                    let item = hir.item(id);
                     match item.kind {
                         rustc_hir::ItemKind::Static(_, _, _) | rustc_hir::ItemKind::Fn(_, _, _) => {
                             let name = item.ident;
-                            let ty = tcx.type_of(tcx.hir().local_def_id(item.hir_id()));
-                            println!("{:?}:\t{:?}", name, ty)
+                            let ty = tcx.type_of(hir.local_def_id(item.hir_id()));
+                            println!("{name:?}:\t{ty:?}")
                         }
                         _ => (),
                     }
