@@ -1,4 +1,5 @@
-#![feature(const_raw_ptr_deref, never_type)]
+// stderr-per-bitwidth
+#![feature(never_type)]
 
 const _: &[!; 1] = unsafe { &*(1_usize as *const [!; 1]) }; //~ ERROR undefined behavior
 const _: &[!; 0] = unsafe { &*(1_usize as *const [!; 0]) }; // ok
