@@ -1,8 +1,11 @@
-// check-pass
+// revisions: base extended
+//[base] check-fail
+//[extended] check-pass
 
 #![feature(generic_associated_types)]
-  //~^ WARNING: the feature `generic_associated_types` is incomplete
 #![feature(associated_type_defaults)]
+#![cfg_attr(extended, feature(generic_associated_types_extended))]
+#![cfg_attr(extended, allow(incomplete_features))]
 
 trait Foo {
     type A<'a> where Self: 'a;
@@ -18,10 +21,11 @@ impl Foo for Fooy {
 struct Fooer<T>(T);
 
 impl<T> Foo for Fooer<T> {
-    type A<'x> where T: 'x = &'x ();
+    type A<'x> = &'x () where T: 'x;
 }
 
 fn f(_arg : Box<dyn for<'a> Foo<A<'a> = &'a ()>>) {}
+//[base]~^ the trait `Foo` cannot be made into an object
 
 
 fn main() {
